from ..bindings.imports import language as wit
from ..csi import Language, SelectLanguageRequest


def language_to_wit(language: Language) -> str:
    return language.value


def language_from_wit(language: str) -> Language:
    try:
        return Language(language)
    except AttributeError:
        raise ValueError(f"Unsupported language: {language}")


def language_request_to_wit(
    request: SelectLanguageRequest,
) -> wit.SelectLanguageRequest:
    return wit.SelectLanguageRequest(
        request.text, [language_to_wit(language) for language in request.languages]
    )



"""
As the bindings for streaming are still behind a feature flag, we can not
require the generated bindings for the corresponding structs (`CompletionAppend`
and `MessageAppend`). We therefore can not use these types for annotations.

The `type: ignore[no-untyped-def]` annotations can be removed once we stabilize
the feature and we know that the classes will always be in the bindings.
"""

from types import TracebackType
from typing import Self

from pharia_skill.csi.inference import (
    ChatEvent,
    ChatStreamResponse,
    CompletionAppend,
    CompletionEvent,
    CompletionStreamResponse,
    MessageAppend,
    MessageBegin,
)

from ..bindings.imports import inference as wit
from ..csi import (
    ChatParams,
    ChatRequest,
    ChatResponse,
    Completion,
    CompletionParams,
    CompletionRequest,
    Distribution,
    ExplanationRequest,
    FinishReason,
    Granularity,
    Logprob,
    Logprobs,
    Message,
    Role,
    TextScore,
    TokenUsage,
    TopLogprobs,
)


class WitCompletionStreamResponse(CompletionStreamResponse):
    def __init__(self, stream: "wit.CompletionStream"):
        self._stream = stream

    def __enter__(self) -> Self:
        self._stream.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        return self._stream.__exit__(exc_type, exc_value, traceback)

    def next(self) -> CompletionEvent | None:
        match self._stream.next():
            case wit.CompletionEvent_Append(value):
                return completion_append_from_wit(value)
            case wit.CompletionEvent_End(value):
                return finish_reason_from_wit(value)
            case wit.CompletionEvent_Usage(value):
                return token_usage_from_wit(value)
            case _:
                return None


class WitChatStreamResponse(ChatStreamResponse):
    def __init__(self, stream: "wit.ChatStream"):
        self._stream = stream
        super().__init__()

    def __enter__(self) -> Self:
        self._stream.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        return self._stream.__exit__(exc_type, exc_value, traceback)

    def next(self) -> ChatEvent | None:
        match self._stream.next():
            case wit.ChatEvent_MessageBegin(value):
                return MessageBegin(value)
            case wit.ChatEvent_MessageAppend(value):
                return message_append_from_wit(value)
            case wit.ChatEvent_MessageEnd(value):
                return finish_reason_from_wit(value)
            case wit.ChatEvent_Usage(value):
                return token_usage_from_wit(value)
            case _:
                return None


def chat_params_to_wit(chat_params: ChatParams) -> wit.ChatParams:
    return wit.ChatParams(
        max_tokens=chat_params.max_tokens,
        temperature=chat_params.temperature,
        top_p=chat_params.top_p,
        frequency_penalty=chat_params.frequency_penalty,
        presence_penalty=chat_params.presence_penalty,
        logprobs=logprobs_to_wit(chat_params.logprobs),
    )


def message_to_wit(message: Message) -> wit.Message:
    return wit.Message(role=message.role.value, content=message.content)


def chat_request_to_wit(chat_request: ChatRequest) -> wit.ChatRequest:
    return wit.ChatRequest(
        model=chat_request.model,
        messages=[message_to_wit(msg) for msg in chat_request.messages],
        params=chat_params_to_wit(chat_request.params),
    )


def logprob_from_wit(logprob: wit.Logprob) -> Logprob:
    return Logprob(token=logprob.token, logprob=logprob.logprob)


def distribution_from_wit(distribution: wit.Distribution) -> Distribution:
    return Distribution(
        sampled=logprob_from_wit(distribution.sampled),
        top=[logprob_from_wit(logprob) for logprob in distribution.top],
    )


def token_usage_from_wit(usage: wit.TokenUsage) -> TokenUsage:
    return TokenUsage(prompt=usage.prompt, completion=usage.completion)


def completion_append_from_wit(append) -> CompletionAppend:  # type: ignore[no-untyped-def]
    return CompletionAppend(
        text=append.text,
        logprobs=[
            distribution_from_wit(distribution) for distribution in append.logprobs
        ],
    )


def message_append_from_wit(append) -> MessageAppend:  # type: ignore[no-untyped-def]
    return MessageAppend(
        content=append.content,
        logprobs=[
            distribution_from_wit(distribution) for distribution in append.logprobs
        ],
    )


def completion_from_wit(completion: wit.Completion) -> Completion:
    return Completion(
        text=completion.text,
        finish_reason=finish_reason_from_wit(completion.finish_reason),
        logprobs=[
            distribution_from_wit(distribution) for distribution in completion.logprobs
        ],
        usage=token_usage_from_wit(completion.usage),
    )


def completion_params_to_wit(
    completion_params: CompletionParams,
) -> wit.CompletionParams:
    return wit.CompletionParams(
        max_tokens=completion_params.max_tokens,
        temperature=completion_params.temperature,
        top_k=completion_params.top_k,
        top_p=completion_params.top_p,
        stop=completion_params.stop,
        return_special_tokens=completion_params.return_special_tokens,
        frequency_penalty=completion_params.frequency_penalty,
        presence_penalty=completion_params.presence_penalty,
        logprobs=logprobs_to_wit(completion_params.logprobs),
    )


def logprobs_to_wit(logprobs: Logprobs) -> wit.Logprobs:
    match logprobs:
        case "no":
            return wit.Logprobs_No()
        case "sampled":
            return wit.Logprobs_Sampled()
        case TopLogprobs():
            return wit.Logprobs_Top(value=logprobs.top)


def completion_request_to_wit(
    completion_request: CompletionRequest,
) -> wit.CompletionRequest:
    return wit.CompletionRequest(
        model=completion_request.model,
        prompt=completion_request.prompt,
        params=completion_params_to_wit(completion_request.params),
    )


def finish_reason_from_wit(reason: wit.FinishReason) -> FinishReason:
    match reason:
        case wit.FinishReason.STOP:
            return FinishReason.STOP
        case wit.FinishReason.LENGTH:
            return FinishReason.LENGTH
        case wit.FinishReason.CONTENT_FILTER:
            return FinishReason.CONTENT_FILTER


def role_from_wit(role: str) -> Role:
    """Convert a wit role (str) to a SDK role (Role).

    While we represent roles in the wit world as strings (no strict typing allows
    for faster iteration), there is value in exposing strongly typed roles in the
    Python SDK. We assume that any unknown role would come with a LLM response
    (after all, the user can only input the variants of the enum), so we default
    to Assistant.
    """
    match role:
        case "user":
            return Role.User
        case "system":
            return Role.System
        case _:
            return Role.Assistant


def message_from_wit(msg: wit.Message) -> Message:
    return Message(role=role_from_wit(msg.role), content=msg.content)


def chat_response_from_wit(response: wit.ChatResponse) -> ChatResponse:
    return ChatResponse(
        message=message_from_wit(response.message),
        finish_reason=finish_reason_from_wit(response.finish_reason),
        logprobs=[
            distribution_from_wit(distribution) for distribution in response.logprobs
        ],
        usage=token_usage_from_wit(response.usage),
    )


def granularity_to_wit(granularity: Granularity) -> wit.Granularity:
    match granularity:
        case Granularity.AUTO:
            return wit.Granularity.AUTO
        case Granularity.SENTENCE:
            return wit.Granularity.SENTENCE
        case Granularity.WORD:
            return wit.Granularity.WORD
        case Granularity.PARAGRAPH:
            return wit.Granularity.PARAGRAPH


def explanation_request_to_wit(
    explanation_request: ExplanationRequest,
) -> wit.ExplanationRequest:
    return wit.ExplanationRequest(
        prompt=explanation_request.prompt,
        target=explanation_request.target,
        model=explanation_request.model,
        granularity=granularity_to_wit(explanation_request.granularity),
    )


def text_score_from_wit(score: wit.TextScore) -> TextScore:
    return TextScore(start=score.start, length=score.length, score=score.score)



"""
This module references bindings which can only be resolved if targeting the `message-stream-skill` world.

Having this module in the import graph when targeting the `skill` world will lead to a build error.
"""

from types import TracebackType
from typing import Self

from pharia_skill.bindings.imports import streaming_output as wit

from .writer import (
    MessageAppend,
    MessageBegin,
    MessageEnd,
    MessageItem,
    MessageWriter,
    Payload,
)


def message_item_to_wit(item: MessageItem[Payload]) -> wit.MessageItem:
    match item:
        case MessageBegin(role):
            attributes = wit.BeginAttributes(role=role)
            return wit.MessageItem_MessageBegin(value=attributes)
        case MessageAppend(text):
            return wit.MessageItem_MessageAppend(value=text)
        case MessageEnd(payload):
            data = payload.model_dump_json().encode() if payload is not None else None
            return wit.MessageItem_MessageEnd(value=data)


class WitMessageWriter(MessageWriter[Payload]):
    def __init__(self, output: wit.StreamOutput):
        self.inner = output

    def __enter__(self) -> Self:
        self.inner.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        return self.inner.__exit__(exc_type, exc_value, traceback)

    def write(self, item: MessageItem[Payload]) -> None:
        message_item = message_item_to_wit(item)
        self.inner.write(message_item)



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref


S = TypeVar('S')
@dataclass
class Some(Generic[S]):
    value: S

T = TypeVar('T')
@dataclass
class Ok(Generic[T]):
    value: T

E = TypeVar('E')
@dataclass(frozen=True)
class Err(Generic[E], Exception):
    value: E

Result = Union[Ok[T], Err[E]]



from typing import Protocol

from .chunking import Chunk, ChunkParams, ChunkRequest
from .document_index import (
    Document,
    DocumentPath,
    IndexPath,
    JsonSerializable,
    SearchFilter,
    SearchRequest,
    SearchResult,
)
from .inference import (
    ChatParams,
    ChatRequest,
    ChatResponse,
    ChatStreamResponse,
    Completion,
    CompletionParams,
    CompletionRequest,
    CompletionStreamResponse,
    ExplanationRequest,
    Granularity,
    Message,
    TextScore,
)
from .language import Language, SelectLanguageRequest


class Csi(Protocol):
    """The Cognitive System Interface (CSI) is a protocol that allows skills to interact with the Pharia Kernel."""

    def completion_stream(
        self, model: str, prompt: str, params: CompletionParams
    ) -> CompletionStreamResponse:
        """Streams completions given a prompt.

        Parameters:
            model (str, required): Name of model to use.
            prompt (str, required): The text to be completed.
            params (CompletionParams, required): Parameters for the requested completion.
        """
        ...

    def chat_stream(
        self, model: str, messages: list[Message], params: ChatParams
    ) -> ChatStreamResponse:
        """Streams chat with a model, where the first item is the role.

        Parameters:
            model (str, required):
                Name of model to use.

            messages (list[Message], required):
                List of messages, alternating between messages from user and assistant.

            params (ChatParams, required):
                Parameters used for the chat.
        """
        ...

    def complete(self, model: str, prompt: str, params: CompletionParams) -> Completion:
        """Generates completions given a prompt.

        Parameters:
            model (str, required): Name of model to use.
            prompt (str, required): The text to be completed.
            params (CompletionParams, required): Parameters for the requested completion.

        Examples::

            prompt = f\"\"\"<|begin_of_text|><|start_header_id|>system<|end_header_id|>

            You are a poet who strictly speaks in haikus.<|eot_id|><|start_header_id|>user<|end_header_id|>

            {input.root}<|eot_id|><|start_header_id|>assistant<|end_header_id|>\"\"\"
            params = CompletionParams(max_tokens=64)
            completion = csi.complete("llama-3.1-8b-instruct", prompt, params)
        """
        request = CompletionRequest(model, prompt, params)
        return self.complete_concurrent([request])[0]

    def complete_concurrent(
        self, requests: list[CompletionRequest]
    ) -> list[Completion]:
        """Generate completions concurrently.

        Parameters:
            requests (list[CompletionRequest], required): List of completion requests.
        """
        ...

    def chunk(self, text: str, params: ChunkParams) -> list[Chunk]:
        """Chunks a text into chunks according to params.

        Parameters:
            text (str, required): Text to be chunked.
            params (ChunkParams, required): Parameter used for chunking, model and maximal number of tokens.

        Examples::

            text = "A very very very long text that can be chunked."
            params = ChunkParams("llama-3.1-8b-instruct", max_tokens=5)
            result = csi.chunk(text, params)
            assert len(result) == 3
        """
        request = ChunkRequest(text, params)
        return self.chunk_concurrent([request])[0]

    def chunk_concurrent(self, requests: list[ChunkRequest]) -> list[list[Chunk]]:
        """Chunk a text into chunks concurrently.

        Parameters:
            requests (list[ChunkRequest], required): List of chunk requests.
        """
        ...

    def chat(
        self, model: str, messages: list[Message], params: ChatParams
    ) -> ChatResponse:
        """Chat with a model.

        Parameters:
            model (str, required):
                Name of model to use.

            messages (list[Message], required):
                List of messages, alternating between messages from user and assistant.

            params (ChatParams, required):
                Parameters used for the chat.

        Examples::

            input = "oat milk"
            msg = Message.user(f\"\"\"You are a poet who strictly speaks in haikus.

            {input}\"\"\")
            model = "llama-3.1-8b-instruct"
            chat_response = csi.chat(model, [msg], ChatParams(max_tokens=64))
        """
        request = ChatRequest(model, messages, params)
        return self.chat_concurrent([request])[0]

    def chat_concurrent(self, requests: list[ChatRequest]) -> list[ChatResponse]:
        """Chat with a model concurrently.

        Parameters:
            requests (list[ChatRequest], required): List of chat requests.
        """
        ...

    def explain(
        self,
        prompt: str,
        target: str,
        model: str,
        granularity: Granularity = Granularity.AUTO,
    ) -> list[TextScore]:
        request = ExplanationRequest(prompt, target, model, granularity)
        return self.explain_concurrent([request])[0]

    def explain_concurrent(
        self, requests: list[ExplanationRequest]
    ) -> list[list[TextScore]]: ...

    def select_language(self, text: str, languages: list[Language]) -> Language | None:
        """Select the detected language for the provided input based on the list of possible languages.

        If no language matches, None is returned.

        Parameters:
            text (str, required): Text input.
            languages (list[Language], required): All languages that should be considered during detection.

        Examples::

            text = "Ich spreche Deutsch nur ein bisschen."
            languages = [Language.English, Language.German]
            result = csi.select_language(text, languages)
        """
        request = SelectLanguageRequest(text, languages)
        return self.select_language_concurrent([request])[0]

    def select_language_concurrent(
        self, requests: list[SelectLanguageRequest]
    ) -> list[Language | None]:
        """Detect the language for multiple texts concurrently.

        Parameters:
            requests (list[SelectLanguageRequest], required): List of select language requests.
        """
        ...

    def search(
        self,
        index_path: IndexPath,
        query: str,
        max_results: int = 1,
        min_score: float | None = None,
        filters: list[SearchFilter] | None = None,
    ) -> list[SearchResult]:
        """Search an existing Index in the Document Index.

        Parameters:
            index_path (IndexPath, required): Index path in the Document Index to access.
            query (str, required): Text to be search for.
            max_results (int, required): Maximal number of results.
            min_score (float, optional, Default NoneNone): Minimal score for result to be included.

        Examples::

            index_path = IndexPath("f13", "wikipedia-de", "luminous-base-asymmetric-64")
            query = "What is the population of Heidelberg?"
            result = csi.search(index_path, query)
            r0 = result[0]
            "Heidelberg" in r0.content, "Heidelberg" in r0.document_path.name # True, True
        """
        request = SearchRequest(
            index_path, query, max_results, min_score, filters or []
        )
        return self.search_concurrent([request])[0]

    def search_concurrent(
        self, requests: list[SearchRequest]
    ) -> list[list[SearchResult]]: ...

    """Execute multiple search requests against the Document Index."""

    def document(self, document_path: DocumentPath) -> Document:
        """Fetch a document from the Document Index.

        Parameters:
            document_path (DocumentPath, required): The document path to get the document from.

        Examples::

            document_path = DocumentPath("f13", "wikipedia-de", "Heidelberg")
            document = csi.document(document_path)
            assert document.path == document_path
        """
        return self.documents([document_path])[0]

    def documents(self, document_paths: list[DocumentPath]) -> list[Document]:
        """Fetch multiple documents from the Document Index.

        Parameters:
            document_paths (list[DocumentPath], required): The document paths to get the documents from.
        """
        ...

    def document_metadata(self, document_path: DocumentPath) -> JsonSerializable:
        """Return the metadata of a document in the Document Index.

        Parameters:
            document_path (DocumentPath, required): The document path to get metadata from.
        """
        return self.documents_metadata([document_path])[0]

    def documents_metadata(
        self, document_paths: list[DocumentPath]
    ) -> list[JsonSerializable]:
        """Return the metadata of multiple documents in the Document Index.

        Parameters:
            document_paths (list[DocumentPath], required): The document paths to get metadata from.
        """
        ...



import logging
import os
import subprocess
import time
from enum import Enum
from typing import Optional

import typer
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
from rich.prompt import Confirm, Prompt
from typing_extensions import Annotated

from .pharia_skill_cli import PhariaSkillCli, Registry

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[logging.StreamHandler()],
)

logger = logging.getLogger(__name__)
console = Console()


def setup_wasi_deps() -> None:
    """Download the Pydantic WASI wheels if they are not already present."""
    PYDANTIC_CORE_VERSION = "2.27.2"
    WASI_DEPS_PATH = "wasi_deps"
    if os.path.exists(WASI_DEPS_PATH):
        if not os.path.exists(
            f"{WASI_DEPS_PATH}/pydantic_core-{PYDANTIC_CORE_VERSION}.dist-info"
        ):
            logger.info("Deleting outdated Pydantic Wasi wheels...")
            subprocess.run(["rm", "-rf", WASI_DEPS_PATH])

    if not os.path.exists(WASI_DEPS_PATH):
        logger.info("Downloading Pydantic Wasi wheels...")
        subprocess.run(
            [
                "pip3",
                "install",
                "--target",
                WASI_DEPS_PATH,
                "--only-binary",
                ":all:",
                "--platform",
                "any",
                "--platform",
                "wasi_0_0_0_wasm32",
                "--python-version",
                "3.12",
                "--index-url",
                "https://benbrandt.github.io/wasi-wheels/",
                "--extra-index-url",
                "https://pypi.org/simple",
                f"pydantic-core=={PYDANTIC_CORE_VERSION}",
            ],
            check=True,
        )


class BuildError(Exception):
    """Any error encountered trying to build the Skill as a WASM component."""

    def __init__(self, message: str):
        self.message = message


class IsMessageStream(BuildError):
    """Skill needs to be built against the `message-stream-skill` world."""

    def __init__(self, message: str):
        self.message = message


class IsSkill(BuildError):
    """Skill is not build against the `skill` world."""

    def __init__(self, message: str):
        self.message = message


class SkillType(str, Enum):
    SKILL = "skill"
    MESSAGE_STREAM_SKILL = "message-stream-skill"


def run_componentize_py(
    skill_module: str, output_file: str, unstable: bool, skill_type: SkillType
) -> str:
    """Build the skill to a WASM component using componentize-py.

    The call to componentize-py targets the `skill` wit world and adds the downloaded
    Pydantic WASI wheels to the Python path.

    Returns:
        str: The path to the generated WASM file.
    """
    setup_wasi_deps()
    args = ["--all-features"] if unstable else []
    command = [
        "componentize-py",
        *args,
        "-w",
        skill_type.value,
        "componentize",
        skill_module,
        "-o",
        output_file,
        "-p",
        ".",
        "-p",
        "wasi_deps",
    ]

    try:
        subprocess.run(
            command,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        if "message_stream" in e.stderr:
            raise IsMessageStream(e.stderr)
        if "skill_handler" in e.stderr:
            raise IsSkill(e.stderr)
        raise BuildError(e.stderr)

    return output_file


def display_publish_suggestion(wasm_file: str) -> None:
    """Display a colorful suggestion to publish the skill.

    Args:
        wasm_file: Path to the generated WASM file.
    """
    wasm_filename = wasm_file.lstrip("./")
    publish_command = f"pharia-skill publish {wasm_filename} --tag [TAG] --name [NAME]"

    console.print(
        Panel.fit(
            f"[bold]Skill:[/bold] [cyan]{wasm_filename}[/cyan]\n\n"
            f"[yellow]To publish, run:[/yellow]\n"
            f"[cyan]{publish_command}[/cyan]",
            title="[bold green]Build Successful[/bold green]",
            border_style="green",
            padding=(1, 2),
        )
    )


def publish_skill(skill_path: str, name: Optional[str], tag: str) -> None:
    """Publish a skill with progress indicator and success message.

    Args:
        skill_path: Path to the WASM file to publish.
        name: Name to publish the skill as, or None to use the filename.
        tag: Tag to publish the skill with.
    """
    if not skill_path.endswith(".wasm"):
        skill_path += ".wasm"

    display_name = name if name else skill_path.replace(".wasm", "")

    cli = PhariaSkillCli()

    try:
        registry = Registry.from_env()
    except KeyError as e:
        console.print(
            Panel(
                f"The environment variable [yellow]{e}[/yellow] is not set.",
                title="[bold red]Error[/bold red]",
                border_style="red",
                padding=(1, 1),
            )
        )
        raise typer.Exit(code=1)

    start_time = time.time()
    with Progress(
        SpinnerColumn(),
        TextColumn(f"Publishing [cyan]{display_name}[/cyan]:[cyan]{tag}[/cyan]..."),
        TimeElapsedColumn(),
        console=console,
        transient=True,
    ) as progress:
        task = progress.add_task("", total=None)
        cli.publish(skill_path, name, tag, registry)
        progress.update(task, completed=True)

    console.print(
        Panel.fit(
            f"[bold]Skill:[/bold] [cyan]{display_name}[/cyan]\n"
            f"[bold]Tag:[/bold] [cyan]{tag}[/cyan]\n"
            f"[bold]Registry:[/bold] [cyan]{registry.registry}/{registry.repository}[/cyan]\n\n"
            f"Published in [yellow]{time.time() - start_time:.2f}[/yellow] seconds",
            title="[bold green]Publish Successful[/bold green]",
            border_style="green",
            padding=(1, 2),
        )
    )


def prompt_for_publish(wasm_file: str) -> None:
    """Prompt the user to publish the skill.

    Args:
        wasm_file: Path to the generated WASM file.
    """
    wasm_filename = wasm_file.lstrip("./")

    # Ask if the user wants to publish now with a more engaging prompt
    if Confirm.ask(
        "\n[bold yellow]Would you like to publish this skill now?[/bold yellow]",
        default=True,
        console=console,
    ):
        tag = Prompt.ask(
            "[bold cyan]Enter tag[/bold cyan]",
            default="latest",
            show_default=True,
            console=console,
        )

        name_default = wasm_filename.replace(".wasm", "")
        name = Prompt.ask(
            "[bold cyan]Enter name[/bold cyan]",
            default=name_default,
            show_default=True,
            console=console,
        )

        # Publish the skill
        publish_skill(wasm_filename, name, tag)


app = typer.Typer(rich_markup_mode="rich")


@app.callback()
def callback() -> None:
    """
    [bold green]Pharia Skill CLI Tool[/bold green].

    A tool for building and publishing Pharia Skills.
    """


@app.command()
def build(
    skill: Annotated[
        str,
        typer.Argument(help="Python module of the skill to build", show_default=False),
    ],
    unstable: Annotated[
        bool,
        typer.Option(
            help="Enable unstable features for testing. Don't try this at home."
        ),
    ] = False,
    interactive: Annotated[
        bool,
        typer.Option(
            help="Prompt for publishing after building.",
            show_default=True,
        ),
    ] = True,
    skill_type: Annotated[
        SkillType,
        typer.Option(
            help="The type of skill to build.",
            show_default=True,
        ),
    ] = SkillType.SKILL,
) -> None:
    """
    [bold blue]Build[/bold blue] a skill.

    Compiles a Python module into a WebAssembly component.
    """
    if "/" in skill or skill.endswith(".py"):
        suggestion = skill
        if skill.endswith(".py"):
            suggestion = skill[:-3]
        if "/" in suggestion:
            suggestion = suggestion.replace("/", ".")

        console.print(
            Panel(
                f"Argument must be a fully qualified Python module name, not [cyan]{skill}[/cyan]\n\n"
                f"[yellow]Did you mean?[/yellow] [green]{suggestion}[/green]\n\n"
                f"[italic]Example: Use [green]my_package.my_module[/green] instead of [red]my_package/my_module.py[/red][/italic]",
                title="[bold red]Error[/bold red]",
                border_style="red",
                padding=(1, 1),
            )
        )
        raise typer.Exit(code=1)

    output_file = f"./{skill.split('.')[-1]}.wasm"
    with Progress(
        SpinnerColumn(),
        TextColumn(
            f"Building WASM component [cyan]{output_file}[/cyan] from module [cyan]{skill}[/cyan]..."
        ),
        TimeElapsedColumn(),
        console=console,
        transient=True,
    ) as progress:
        task = progress.add_task("", total=None)
        try:
            wasm_file = run_componentize_py(skill, output_file, unstable, skill_type)
            progress.update(task, completed=True)
        except IsMessageStream:
            console.print(
                Panel(
                    "It seems you are trying to build a Skill with the @message_stream decorator.\nPlease ensure to set the --skill-type flag to [green]message-stream-skill[/green].",
                    title="[bold red]Error[/bold red]",
                )
            )
            raise typer.Exit(code=1)
        except IsSkill:
            console.print(
                Panel(
                    "It seems you are trying to build a Skill decorated with the @skill decorator.\nPlease ensure to set the --skill-type flag to [green]skill[/green].",
                    title="[bold red]Error[/bold red]",
                )
            )
            raise typer.Exit(code=1)
    if wasm_file and interactive:
        display_publish_suggestion(wasm_file)
        prompt_for_publish(wasm_file)


@app.command()
def publish(
    skill: Annotated[
        str,
        typer.Argument(
            help="A path to a Wasm file containing a Skill.", show_default=False
        ),
    ],
    name: Annotated[
        Optional[str],
        typer.Option(
            help="The name to publish the Skill as. If not provided, it is inferred based on the Wasm filename.",
            show_default="The filename",
        ),
    ] = None,
    tag: Annotated[str, typer.Option(help="An identifier for the Skill.")] = "latest",
) -> None:
    """
    [bold blue]Publish[/bold blue] a skill.

    Publishes a WebAssembly component to the Pharia Skill registry.
    """
    publish_skill(skill, name, tag)


if __name__ == "__main__":
    app()



from pydantic import RootModel

from pharia_skill.csi import (
    Document,
    DocumentPath,
    JsonSerializable,
    SearchRequest,
    SearchResult,
)

DocumentMetadataSerializer = RootModel[list[DocumentPath]]


DocumentMetadataDeserializer = RootModel[list[JsonSerializable | None]]


DocumentSerializer = RootModel[list[DocumentPath]]


DocumentDeserializer = RootModel[list[Document]]


SearchRequestSerializer = RootModel[list[SearchRequest]]


SearchResultDeserializer = RootModel[list[list[SearchResult]]]



"""
The `ChatRequest` is the entire conversation state and its context.

It is initially constructed by the user by providing a user and potentially a
system message. Along with the model and completion parameters, also tools that
are available to the model (context) can be specified.

The `ChatRequest` can be extended from the LLM side by passing it to the
`chat` function. Afterwards, it can be extended from the user side by adding
a message or tool response to the request.
"""

from dataclasses import dataclass, field
from typing import Any, Literal, Sequence

from pydantic import (
    BaseModel,
    field_serializer,
    field_validator,
)

from pharia_skill.csi import ChatParams, CompletionParams, Csi, FinishReason

from .message import (
    AssistantMessage,
    Role,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from .response import SpecialTokens
from .tool import BuiltInTools, JsonSchema, ToolDefinition

Message = UserMessage | AssistantMessage | ToolMessage


class BuiltInToolSchema(BaseModel):
    """Built-in tools that can be specified."""

    type: Literal["code_interpreter", "wolfram_alpha", "brave_search"]


@dataclass
class ChatResponse:
    """Response from a chat request."""

    message: AssistantMessage
    finish_reason: FinishReason


@dataclass
class ChatRequest:
    """Represents the conversation state and context.

    Conversation history and available tools can be provided on initialization.
    Is automatically extended when passing it to the `chat` function.
    Can be reused for subsequent conversation turns by extending it with messages
    or tool results and passing it to the `chat` function again.
    """

    model: str
    messages: list[Message]
    """The message history to be passed to the model. Can be `User`, `Assistant`, or `Tool` messages.
    
    Note that if you want to include a system prompt, you can use the top-level system parameter —
    there is no `system` role for input messages in the Messages API."""

    system: str | None = None
    """An optional system message that can passed to the model.

    Note that you don't have to define tools or specify `Environment: ipython` if you are using tools.
    The system message is automatically adapted.
    """

    tools: Sequence[ToolDefinition] = field(default_factory=list)
    params: ChatParams = field(default_factory=ChatParams)

    def __post_init__(self) -> None:
        validate_messages(self.messages)

    def chat(self, csi: Csi) -> ChatResponse:
        """Chat with a Llama model.

        Appends responses from the model to the conversation history.
        Available tools can be specified as part of the `ChatRequest`. If the model decides
        to do a tool call, this will be available on the response:

        Example::
            # define the tool
            class GetGithubReadme(BaseModel):
                repository: str

            # construct the `ChatRequest` with the user question
            user = Message.user("When will my order (42) arrive?")
            request = ChatRequest(llama, [user], tools=[GetGithubReadme])

            # chat with the model
            response = request.chat(csi)

            # receive the tool call back
            assert response.message.tool_call is not None

            # execute the tool call (e.g. via http request) and construct the tool response
            tool_response = ToolResponse(tool.name, content="1970-01-01")

            # extend the request and run a new chat
            request.extend(tool_response)
            response = request.chat(csi)
        """
        validate_messages(self.messages)
        completion_params = to_completion_params(self.params)
        completion = csi.complete(self.model, self.render(), completion_params)
        message = AssistantMessage.from_raw_response(completion.text, self.tools)

        self.messages.append(message)
        return ChatResponse(message, completion.finish_reason)

    def extend(self, message: Message) -> None:
        """Add a message to a chat request.

        When conversations have multiple turns (e.g. with tool calling), the
        conversation can be extended by adding a message to the request.
        """
        validate_messages(self.messages + [message])
        self.messages.append(message)

    def render(self) -> str:
        """Convert the chat request to a prompt that can be passed to the model."""
        prompt = SpecialTokens.BeginOfText.value
        if self.system or self.tools:
            prompt += SystemMessage(self.system or "").render(self.tools)

        # tools only get passed to the first user message
        for message in self.messages:
            prompt += message.render([])

        prompt += Role.Assistant.render()
        prompt += "\n\n"
        return prompt

    @field_serializer("tools")
    def as_dict(self, tools: Sequence[ToolDefinition]) -> list[dict[str, Any]]:
        """Pydantic can not serialize type[BaseModel], so we serialize it manually."""
        return [
            t.model_dump() if isinstance(t, JsonSchema) else t.json_schema()
            for t in tools
        ]

    @field_validator(
        "tools",
        mode="before",
        json_schema_input_type=list[JsonSchema | BuiltInToolSchema],
    )
    @classmethod
    def validate_tools(cls, value: Any) -> Sequence[ToolDefinition]:
        assert isinstance(value, list)
        tools = []
        for tool in value:
            assert isinstance(tool, dict) and tool.get("type")
            if tool["type"] != "function":
                known = next(
                    (b for b in BuiltInTools if tool["type"] == b.name()), None
                )
                if not known:
                    raise ValueError(f"Invalid built in tool: {tool}")
                tools.append(known)
            elif isinstance(tool, dict):
                tools.append(JsonSchema(**tool))  # type: ignore
        return tools


def to_completion_params(params: ChatParams) -> CompletionParams:
    return CompletionParams(
        return_special_tokens=True,
        max_tokens=params.max_tokens,
        temperature=params.temperature,
        top_p=params.top_p,
        stop=[SpecialTokens.StartHeader.value],
    )


def validate_messages(messages: list[Message]) -> None:
    """Validate the order of messages in a chat request.

    We enforce a sensible Llama3 prompt (even though it might not be required by the model),
    to give the developer early feedback if he is using the chat request incorrectly.

    Rules:
        1. The first message must be a user message.
        2. The conversation must alternate between user/ipython and assistant.
        3. The last message must be a user or ipython message.
    """
    if not messages:
        raise ValueError("Messages cannot be empty")

    # 1. check that the first message is a system or user message
    if messages[0].role != Role.User:
        raise ValueError("First message must be a user message")

    # 2. check alternating between user/tool and assistant
    for i, message in enumerate(messages[1:]):
        if i % 2 == 0:
            if message.role != Role.Assistant:
                raise ValueError("Assistant messages must alternate with user messages")
        else:
            if message.role not in (Role.User, Role.IPython):
                raise ValueError("User messages must alternate with assistant messages")

    # 3. check that the last message is a user/ipython message
    if messages[-1].role not in (Role.User, Role.IPython):
        raise ValueError("Last message must be a user or ipython message")



import datetime as dt
import json
from typing import cast

from ..bindings.imports import document_index as wit
from ..csi import (
    After,
    AtOrAfter,
    AtOrBefore,
    Before,
    Cursor,
    Document,
    DocumentPath,
    EqualTo,
    FilterCondition,
    GreaterThan,
    GreaterThanOrEqualTo,
    Image,
    IndexPath,
    IsNull,
    JsonSerializable,
    LessThan,
    LessThanOrEqualTo,
    MetadataFilter,
    Modality,
    SearchFilter,
    SearchRequest,
    SearchResult,
    Text,
    With,
    WithOneOf,
    Without,
)


def to_isostring(datetime: dt.datetime) -> str:
    """The WIT world represents instants as ISO 8601 strings.

    While the Document Index supports loading from any timezoned string, we do require specifying as timezone.
    """
    assert datetime.tzinfo is not None, "Datetimes must be timezone-aware"
    return datetime.isoformat()


def value_to_wit(value: str | int | bool) -> wit.MetadataFieldValue:
    match value:
        case str():
            return wit.MetadataFieldValue_StringType(value)
        case int():
            return wit.MetadataFieldValue_IntegerType(value)
        case bool():
            return wit.MetadataFieldValue_BooleanType(value)


def condition_to_wit(condition: FilterCondition) -> wit.MetadataFilterCondition:
    match condition:
        case GreaterThan(value):
            return wit.MetadataFilterCondition_GreaterThan(value)
        case GreaterThanOrEqualTo(value):
            return wit.MetadataFilterCondition_GreaterThanOrEqualTo(value)
        case LessThan(value):
            return wit.MetadataFilterCondition_LessThan(value)
        case LessThanOrEqualTo(value):
            return wit.MetadataFilterCondition_LessThanOrEqualTo(value)
        case After(value):
            return wit.MetadataFilterCondition_After(to_isostring(value))
        case AtOrAfter(value):
            return wit.MetadataFilterCondition_AtOrAfter(to_isostring(value))
        case Before(value):
            return wit.MetadataFilterCondition_Before(to_isostring(value))
        case AtOrBefore(value):
            return wit.MetadataFilterCondition_AtOrBefore(to_isostring(value))
        case EqualTo(value):
            return wit.MetadataFilterCondition_EqualTo(value_to_wit(value))
        case IsNull():
            return wit.MetadataFilterCondition_IsNull()


def metadata_filter_to_wit(filter: MetadataFilter) -> wit.MetadataFilter:
    return wit.MetadataFilter(
        field=filter.field,
        condition=condition_to_wit(filter.condition),
    )


def filter_to_wit(filter: SearchFilter) -> wit.SearchFilter:
    match filter:
        case Without(value):
            return wit.SearchFilter_Without(
                value=[metadata_filter_to_wit(f) for f in value]
            )
        case WithOneOf(value):
            return wit.SearchFilter_WithOneOf(
                value=[metadata_filter_to_wit(f) for f in value]
            )
        case With(value):
            return wit.SearchFilter_WithAll(
                value=[metadata_filter_to_wit(f) for f in value]
            )


def search_request_to_wit(request: SearchRequest) -> wit.SearchRequest:
    return wit.SearchRequest(
        index_path=index_path_to_wit(request.index_path),
        query=request.query,
        max_results=request.max_results,
        min_score=request.min_score,
        filters=[filter_to_wit(f) for f in request.filters],
    )


def document_path_from_wit(document_path: wit.DocumentPath) -> DocumentPath:
    return DocumentPath(
        namespace=document_path.namespace,
        collection=document_path.collection,
        name=document_path.name,
    )


def cursor_from_wit(cursor: wit.TextCursor) -> Cursor:
    return Cursor(item=cursor.item, position=cursor.position)


def search_result_from_wit(result: wit.SearchResult) -> SearchResult:
    return SearchResult(
        document_path=document_path_from_wit(result.document_path),
        content=result.content,
        score=result.score,
        start=cursor_from_wit(result.start),
        end=cursor_from_wit(result.end),
    )


def document_path_to_wit(document_path: DocumentPath) -> wit.DocumentPath:
    return wit.DocumentPath(
        namespace=document_path.namespace,
        collection=document_path.collection,
        name=document_path.name,
    )


def document_contents_from_wit(contents: list[wit.Modality]) -> list[Modality]:
    return [
        Text(content.value) if isinstance(content, wit.Modality_Text) else Image()
        for content in contents
    ]


def document_metadata_from_wit(metadata: bytes | None) -> JsonSerializable:
    return cast(JsonSerializable, json.loads(metadata)) if metadata else None


def document_from_wit(document: wit.Document) -> Document:
    metadata = document_metadata_from_wit(document.metadata)
    contents = document_contents_from_wit(document.contents)
    path = document_path_from_wit(document.path)
    return Document(path=path, contents=contents, metadata=metadata)


def index_path_to_wit(index_path: IndexPath) -> wit.IndexPath:
    return wit.IndexPath(
        namespace=index_path.namespace,
        collection=index_path.collection,
        index=index_path.index,
    )



from dataclasses import dataclass
from typing import Generic

from pharia_skill.message_stream.writer import (
    MessageAppend,
    MessageBegin,
    MessageEnd,
    MessageItem,
    MessageWriter,
    Payload,
)


@dataclass
class RecordedMessage(Generic[Payload]):
    role: str | None
    content: str = ""
    payload: Payload | None = None


class MessageRecorder(MessageWriter[Payload]):
    """A message writer that can be passed into a `message_stream` skill at testing time.

    It allows to inspect the output that a skill produces, either via the `items` property
    that stored individual chunks that have been written or via the `messages` method
    that aggregates the items into a list of messages.

    The MessageRecorder also validates the stream of items that are written to it.

    Example::

        from pharia_skill import Csi, message_stream, MessageAppend, MessageBegin, MessageEnd
        from pharia_skill.testing import MessageWriter, MessageRecorder, RecordedMessage

        @message_stream
        def my_skill(csi: Csi, writer: MessageWriter, input: Input) -> None:
            ...

        def test_my_skill():
            csi = DevCsi()
            writer = MessageRecorder()
            input = Input(topic="The meaning of life")

            my_skill(csi, writer, input)

            assert writer.messages() == [
                RecordedMessage(role="assistant", content="The meaning of life"),
            ]
    """

    def __init__(self) -> None:
        self.items: list[MessageItem[Payload]] = []

    def write(self, item: MessageItem[Payload]) -> None:
        """Store and validate the streamed items.

        Validating the stream here gives the developer early feedback at test time.
        """
        MessageRecorder.validate(self.items, item)
        self.items.append(item)

    @staticmethod
    def validate(
        existing: list[MessageItem[Payload]], item: MessageItem[Payload]
    ) -> None:
        """Is it legal to append this item to the previous items?

        There are three rules that must be followed:

        1. The first item must be a `MessageBegin`.
        2. Consecutive `MessageBegin`s must be preceded by a `MessageEnd`.
        3. A `MessageEnd` must not be preceded by `MessageEnd`.
        """

        if not existing:
            if not isinstance(item, MessageBegin):
                raise ValueError("The first item must be a `MessageBegin`")
            return

        if isinstance(item, MessageBegin) and not isinstance(existing[-1], MessageEnd):
            raise ValueError(
                "Consecutive `MessageBegin`s must be preceded by a `MessageEnd`"
            )

        if isinstance(item, MessageEnd) and isinstance(existing[-1], MessageEnd):
            raise ValueError(
                "A `MessageEnd` must not be preceded by another `MessageEnd`"
            )

    def messages(self) -> list[RecordedMessage[Payload]]:
        """Convenience method to aggregate the streamed items into a list of messages.

        Message items are validated when they are written, so we assume that the list is
        valid.
        """
        messages: list[RecordedMessage[Payload]] = []
        for item in self.items:
            match item:
                case MessageBegin(role=role):
                    messages.append(RecordedMessage(role=role))
                case MessageAppend(text=text):
                    messages[-1].content += text
                case MessageEnd(payload=payload):
                    messages[-1].payload = payload
        return messages



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from ..types import Result, Ok, Err, Some


@dataclass
class IndexPath:
    """
    Which documents you want to search in, and which type of index should be used
    """
    namespace: str
    collection: str
    index: str

@dataclass
class DocumentPath:
    namespace: str
    collection: str
    name: str

@dataclass
class TextCursor:
    """
    A position within a document. The cursor is always inclusive of the current position, in both start and end positions.
    """
    item: int
    position: int

@dataclass
class SearchResult:
    """
    The result for semantic document search. Part of an array of document names and content of the found documents in the given collection.
    """
    document_path: DocumentPath
    content: str
    score: float
    start: TextCursor
    end: TextCursor


@dataclass
class MetadataFieldValue_StringType:
    value: str


@dataclass
class MetadataFieldValue_IntegerType:
    value: int


@dataclass
class MetadataFieldValue_BooleanType:
    value: bool


MetadataFieldValue = Union[MetadataFieldValue_StringType, MetadataFieldValue_IntegerType, MetadataFieldValue_BooleanType]



@dataclass
class MetadataFilterCondition_GreaterThan:
    value: float


@dataclass
class MetadataFilterCondition_GreaterThanOrEqualTo:
    value: float


@dataclass
class MetadataFilterCondition_LessThan:
    value: float


@dataclass
class MetadataFilterCondition_LessThanOrEqualTo:
    value: float


@dataclass
class MetadataFilterCondition_After:
    value: str


@dataclass
class MetadataFilterCondition_AtOrAfter:
    value: str


@dataclass
class MetadataFilterCondition_Before:
    value: str


@dataclass
class MetadataFilterCondition_AtOrBefore:
    value: str


@dataclass
class MetadataFilterCondition_EqualTo:
    value: MetadataFieldValue


@dataclass
class MetadataFilterCondition_IsNull:
    pass


MetadataFilterCondition = Union[MetadataFilterCondition_GreaterThan, MetadataFilterCondition_GreaterThanOrEqualTo, MetadataFilterCondition_LessThan, MetadataFilterCondition_LessThanOrEqualTo, MetadataFilterCondition_After, MetadataFilterCondition_AtOrAfter, MetadataFilterCondition_Before, MetadataFilterCondition_AtOrBefore, MetadataFilterCondition_EqualTo, MetadataFilterCondition_IsNull]


@dataclass
class MetadataFilter:
    """
    Matches sections whose metadata fields match the given condition. You must specify the field, and can only specify a single condition.
    """
    field: str
    condition: MetadataFilterCondition


@dataclass
class SearchFilter_Without:
    value: List[MetadataFilter]


@dataclass
class SearchFilter_WithOneOf:
    value: List[MetadataFilter]


@dataclass
class SearchFilter_WithAll:
    value: List[MetadataFilter]


SearchFilter = Union[SearchFilter_Without, SearchFilter_WithOneOf, SearchFilter_WithAll]
"""
A logical combination of filter conditions.
"""


@dataclass
class SearchRequest:
    index_path: IndexPath
    query: str
    max_results: int
    min_score: Optional[float]
    filters: List[SearchFilter]


@dataclass
class Modality_Text:
    value: str


@dataclass
class Modality_Image:
    pass


Modality = Union[Modality_Text, Modality_Image]


@dataclass
class Document:
    path: DocumentPath
    contents: List[Modality]
    metadata: Optional[bytes]


def search(requests: List[SearchRequest]) -> List[List[SearchResult]]:
    raise NotImplementedError

def document_metadata(requests: List[DocumentPath]) -> List[Optional[bytes]]:
    raise NotImplementedError

def documents(requests: List[DocumentPath]) -> List[Document]:
    raise NotImplementedError




from .csi import (
    After,
    AtOrAfter,
    AtOrBefore,
    Before,
    ChatParams,
    ChatRequest,
    ChatResponse,
    Chunk,
    ChunkParams,
    ChunkRequest,
    Completion,
    CompletionParams,
    CompletionRequest,
    Csi,
    Cursor,
    Distribution,
    Document,
    DocumentPath,
    EqualTo,
    ExplanationRequest,
    FilterCondition,
    FinishReason,
    Granularity,
    GreaterThan,
    GreaterThanOrEqualTo,
    Image,
    IndexPath,
    IsNull,
    JsonSerializable,
    Language,
    LessThan,
    LessThanOrEqualTo,
    Logprob,
    Logprobs,
    Message,
    MetadataFilter,
    Modality,
    NoLogprobs,
    Role,
    SampledLogprobs,
    SearchFilter,
    SearchRequest,
    SearchResult,
    SelectLanguageRequest,
    Text,
    TextScore,
    TokenUsage,
    TopLogprobs,
    With,
    WithOneOf,
    Without,
)
from .message_stream import (
    MessageAppend,
    MessageBegin,
    MessageEnd,
    MessageItem,
    MessageWriter,
    message_stream,
)
from .skill import skill

__all__ = [
    "After",
    "AtOrAfter",
    "AtOrBefore",
    "Before",
    "ChatParams",
    "ChatRequest",
    "ChatResponse",
    "Chunk",
    "ChunkParams",
    "ChunkRequest",
    "Completion",
    "CompletionParams",
    "CompletionRequest",
    "Csi",
    "Cursor",
    "Distribution",
    "Document",
    "DocumentPath",
    "EqualTo",
    "ExplanationRequest",
    "FinishReason",
    "Granularity",
    "GreaterThan",
    "GreaterThanOrEqualTo",
    "Image",
    "IndexPath",
    "IsNull",
    "JsonSerializable",
    "Language",
    "LessThan",
    "LessThanOrEqualTo",
    "Logprob",
    "Logprobs",
    "Message",
    "MessageAppend",
    "MessageBegin",
    "MessageEnd",
    "MessageItem",
    "message_stream",
    "MetadataFilter",
    "FilterCondition",
    "Modality",
    "NoLogprobs",
    "Role",
    "MessageWriter",
    "SampledLogprobs",
    "SearchFilter",
    "SearchRequest",
    "SearchResult",
    "SelectLanguageRequest",
    "skill",
    "Text",
    "TextScore",
    "TokenUsage",
    "TopLogprobs",
    "Without",
    "With",
    "WithOneOf",
]



"""
Make HTTP requests against a running Pharia Kernel.

By separating the client from the `DevCsi`, we can better test the serialization/deserialization
and other functionality of the `DevCsi` without making HTTP requests.
"""

import os
from http import HTTPStatus
from typing import Any, Generator, Protocol

import requests
from dotenv import load_dotenv
from sseclient import Event, SSEClient


class CsiClient(Protocol):
    def run(self, function: str, data: dict[str, Any]) -> Any: ...
    def stream(
        self, function: str, data: dict[str, Any]
    ) -> Generator[Event, None, None]: ...


class Client(CsiClient):
    """Make requests with a given payload against a running Pharia Kernel."""

    HTTP_CSI_VERSION = "v1"

    def __init__(self) -> None:
        """Create a new HTTP client.

        The session is stored to allow for re-use of the same connection between tests.
        """
        load_dotenv()
        self.url = (
            f"""{os.environ["PHARIA_KERNEL_ADDRESS"]}/csi/{self.HTTP_CSI_VERSION}"""
        )
        token = os.environ["PHARIA_AI_TOKEN"]
        self.session = requests.Session()
        self.session.headers = {
            "Authorization": f"Bearer {token}",
        }

    def __del__(self) -> None:
        if hasattr(self, "session"):
            self.session.close()

    def run(self, function: str, data: Any) -> Any:
        url = f"{self.url}/{function}"
        response = self.session.post(
            url,
            json=data,
        )
        # Always forward the error message from the kernel
        if response.status_code >= 400:
            try:
                error = response.json()
            except requests.JSONDecodeError:
                error = response.text
            raise Exception(
                f"{response.status_code} {HTTPStatus(response.status_code).phrase}: {error}"
            )

        return response.json()

    def stream(
        self, function: str, data: dict[str, Any]
    ) -> Generator[Event, None, None]:
        url = f"{self.url}/{function}"
        headers = {"Accept": "text/event-stream", **self.session.headers}
        response = self.session.post(url, json=data, headers=headers, stream=True)
        # Always forward the error message from the kernel
        if response.status_code >= 400:
            try:
                error = response.json()
            except requests.JSONDecodeError:
                error = response.text
            raise Exception(
                f"{response.status_code} {HTTPStatus(response.status_code).phrase}: {error}"
            )

        return SSEClient(response).events()  # type: ignore



"""
Provided host types for supporting running streaming skills.
"""
from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from ..types import Result, Ok, Err, Some


@dataclass
class BeginAttributes:
    """
    Payload for the beginning of a message.
    """
    role: Optional[str]


@dataclass
class MessageItem_MessageBegin:
    value: BeginAttributes


@dataclass
class MessageItem_MessageAppend:
    value: str


@dataclass
class MessageItem_MessageEnd:
    value: Optional[bytes]


MessageItem = Union[MessageItem_MessageBegin, MessageItem_MessageAppend, MessageItem_MessageEnd]


class StreamOutput:
    
    def write(self, item: MessageItem) -> None:
        """
        Puts part of a message into the output stream.
        """
        raise NotImplementedError
    def __enter__(self) -> Self:
        """Returns self"""
        return self
                                
    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None) -> bool | None:
        """
        Release this resource.
        """
        raise NotImplementedError



